;Spc2Midi - DSP Emulation (not really)
;============================================================

DspReg:
; DSP special registers, all are both read and write
; (the following apply to each of the 8 channels)
.VolLeft                equ 0   ;       VOLL volume of left channel, 8bits signed
.VolRight               equ 1   ;       VOLR volume of right channel, 8bits signed
.PitchLow               equ 2   ;       PL low byte of 14bit sample pitch
.PitchHigh              equ 3   ;       PH high byte, 1000h is 32khz
.SourceNumber           equ 4   ;       SRCN sample to play 0-255
.AttackDecay            equ 5   ;       Attack and decay rate
.Sustain                equ 6   ;       Sustain level and rate
.Gain                   equ 7   ;       GAIN ??
  .Gain_Amount          equ 224 ;       bits 0-4
  .Gain_Type            equ 31  ;       bits 5-7
  .Gain_Direct          equ 0
  .Gain_IncLinear       equ 128+64
  .Gain_IncBentLine     equ 128+64+32
  .Gain_DecLinear       equ 128+0
  .Gain_DecExponent     equ 128+32
.Envelope               equ 8   ; r     ENVX
.Outx                   equ 9   ; r     ??
.Filter                 equ 0Fh ;
; (these apply to all channels)
.MainVolLeft            equ 0Ch ;       MVOL L
.MainVolRight           equ 1Ch ;       MVOL R
.EchoVolLeft            equ 2Ch ;       EVOL L
.EchoVolRight           equ 3Ch ;       EVOL R
.KeyOn                  equ 4Ch ;       KON voices 0-7, turn on
.KeyOff                 equ 5Ch ;       KOF voices 0-7, mute
.Flags                  equ 6Ch ;       FLG
  .Flags_SoftReset      equ 128
  .Flags_MuteAll        equ 64  ;       turns off all voices when true
  .Flags_Ecen           equ 32  ;       ECEN external memory echo ??
  .Flags_NoiseClock     equ 31  ;       NCK bits 0-4, noise generator clock
.Endx                   equ 7Ch ;       ENDX voices 0-7, flag voice sample has ended
.EchoFeedBack           equ 0Dh ;       EFB voices 0-7
;.Unused                equ 1Dh ;       --
.PitchModulation        equ 2Dh ;       PMOD voices 1-7 (voice 0 not affected)
.NoiseOnOff             equ 3Dh ;       NON voices 0-7
.EchoOnOff              equ 4Dh ;       EON voices 0-7
.SourceDirectory        equ 5Dh ;       DIR address of source directory * 256
.EchoDirectory          equ 6Dh ;       ESA address of echo region * 256 ??
.EchoDelay              equ 7Dh ;       EDL bits 0-3 ??


section .text
;==============================
DspEmu:
;
; What each register holds during emulation:
;
; al  = emulated flags
; ah  = temp flags, otherwise free
; bl  = A
; bh  = Y
; ax+ = cycle counter          (upper 16 bits)
; bx+ = instruction counter    (upper 16 bits)
; ecx = free for opcodes
; ch  = used byte read/write to return and set bytes
; dx  = free for opcode pointer use   (upper 16 bits must remain zero!)
; si  = current opcode pointer (upper 16 bits must remain zero!)
; edi = free for opcodes and register functions
; ebp = free for opcodes and register functions
;
; X, direct page, and stack pointer are not kept in registers.
;
; Opcode routines may use ecx,edx,edi,ah,ebp.
; Read/write routines may only use ah,edi,ebp. All others must be saved.
; edx must be reset to F3h before returning (address of the dsp data
; register) before returning.
; Some instructions require the use of specific registers, such as cl for
; shifting and ax for multiplying. For these, the registers may be saved
; (push) or temporarily swapped with another (xchg).

;==============================
; (edx=register read/written, ch=value read/written)
.wReg03:;pitch, high 6bits of 14bit pitch, 1000h is 32khz
        and ch,3Fh
.wReg00:
.wReg01:
.wReg02:;pitch, low 8bits of 14bit pitch, 1000h is 32khz
.wReg04:;sample source number
.wReg05:
.wReg06:
.wReg07:
.wReg08:
.wReg09:
.wReg0A:
.wReg0B:
.wReg0C:
.wReg0D:
.wReg0E:
.wReg0F:
    mov [DspRam+edx],ch             ;save new value to register
.wIgnore:
.wReg7C:                                ;ignore writes to endx bits
    mov edx,0F3h                    ;restore DSP data address
    ret

;====================
.wReg4C:;key on
    test ch,ch                      ;a lot of games send pointless commands
    jz .wIgnore                     ;why do games turn on zero keys?
    mov ah,ch
    not ah
    or [DspRam+DspReg.KeyOn],ch     ;set new keys on
    and [DspRam+DspReg.KeyOff],ah   ;clear corresponding keys off

   ;Get cycle time
    mov ebp,eax                 ;copy cycles completed
    sar ebp,16                  ;cycles are in upper 16 bits
    add ebp,[SpcEmu.NextTime]   ;get difference of cycles before next event and cycles actually emulated

   ;Record write in note buffer
    push ebx
    mov edi,[NoteBufferEnd]     ;get last position in note buffer
    xor edx,edx
    mov ah,ch                   ;which voices to turn on
    jmp short .CheckNextKeyOn
.NextKeyOn:
    mov [NoteBuffer+edi],ebp    ;save cycle time when note was recorded
    mov bl,dl                   ;voice number (lowest four bits are zero)
    mov [NoteBuffer+edi+4],bl   ;indicate key on
    movzx ebx,word [DspRam+edx+DspReg.PitchLow];get pitch
    mov bl,[PitchTable+ebx]     ;get note half-semitone
    sub bl,46                   ;!!!
    mov [NoteBuffer+edi+5],bl   ;set note number (half-semitone)
    movzx ebx,byte [DspRam+edx+DspReg.SourceNumber] ;get channel sample number
    mov [NoteBuffer+edi+6],bl   ;set sample number
    bts [SampleInfoTable.Used],ebx
  ;%if SpcDebugIoPrint = -1
  ;  jc .SampleAlreadyUsed
  ;  pusha
  ;  mov [esp+StackReg.Cx+1],bl  ;sample used
  ;  mov edi,Text.SampleTest
  ;  call SpcEmu.PrintBreakWithValue
  ;  popa
  ;.SampleAlreadyUsed:;!!!
  ;%endif
    mov bl,[DspRam+edx+DspReg.VolLeft] ;get channel volume
    mov [NoteBuffer+edi+7],al   ;set volume
    add edi,byte 8              ;next position in buffer
    and edi,NoteBufferSize-1
    add edx,byte 16
.CheckNextKeyOn:
    shr ah,1
    jc .NextKeyOn
    lea edx,[edx+16]
    jnz .CheckNextKeyOn
    mov [NoteBufferEnd],edi     ;set last position in event buffer
    pop ebx

   ;Record write in DSP buffer
    mov edi,[DspBufferPointer]  ;get position in DSP buffer
    mov edx,2                   ;start with pitch
    mov ah,ch                   ;which voices to turn on
    jmp short .CheckNextKeyOn2
.NextKeyOn2:
    call .RecordDspFrom
    inc edx
    call .RecordDspFrom
    inc edx
    call .RecordDspFrom
    add edx,byte 14
.CheckNextKeyOn2:
    shr ah,1
    jc .NextKeyOn2
    lea edx,[edx+16]
    jnz .CheckNextKeyOn2
    mov edx,4Ch                 ;restore key_on address
    jmp .RecordDspAfterInit

;====================
.wReg5C:;key off
    mov ah,[DspRam+DspReg.KeyOn];compare channels currently on
    and ah,ch                   ;with channels to turn off
    jz near .wIgnore            ;skip if command affects nothing
    xor [DspRam+DspReg.KeyOn],ah;clear any channels turned off
    or [DspRam+DspReg.KeyOff],ch;set voices as off

   ;Get cycle time
    mov ebp,eax                 ;copy cycles completed
    sar ebp,16                  ;cycles are in upper 16 bits
    add ebp,[SpcEmu.NextTime]   ;get difference of cycles before next event and cycles actually emulated

   ;Record write in note buffer
    mov edi,[NoteBufferEnd]     ;get last position in event buffer
    xor edx,edx                 ;start with first voice
    mov ah,ch                   ;copy which voices to turn off
    jmp short .CheckNextKeyOff
.NextKeyOff:
    mov [NoteBuffer+edi],ebp    ;save cycle time when note off was recorded
    mov dh,dl
    inc dh                      ;same as OR 1
    mov [NoteBuffer+edi+4],dh   ;indicate key off
    add edi,8                   ;next position in buffer
    and edi,NoteBufferSize-1
    add edx,byte 16
.CheckNextKeyOff:
    shr ah,1
    jc .NextKeyOff
    lea edx,[edx+16]
    jnz .CheckNextKeyOff
    mov [NoteBufferEnd],edi     ;set last position in event buffer

    mov edi,[DspBufferPointer]  ;get position in event buffer
    mov edx,5Ch                 ;restore key_off address
    jmp short .RecordDspAfterInit

.wRegXX:
    pusha
%if SpcDebugMode != 1
	mov [SpcEmu.AY],bx
	mov [SpcEmu.Pc],si
	mov [SpcEmu.Flags],al
    mov edi,MakeOperandString.Buffer
	call MakeOperandString
    mov edx,MakeOperandString.Buffer
	call WriteString
%endif
    mov edi,Text.DspWriteIllegal
    call SpcEmu.PrintBreakWithValue
    popa
    ret

;====================
; (edx=dsp register, ch=value written) (edi=buffer ptr, ebp=cycle time)
.RecordDsp:
    mov ebp,eax                 ;copy cycles completed
    sar ebp,16                  ;cycles are in upper 16 bits
    mov edi,[DspBufferPointer]  ;get position in event buffer
    add ebp,[SpcEmu.NextTime]   ;get difference of cycles before next event and cycles actually emulated

;====================
; (edx=dsp register, ch=value written, edi=buffer ptr, ebp=cycle time)
.RecordDspAfterInit:
    mov dh,ch                   ;copy value sent to register
    mov [DspEventBuffer+edi],ebp    ;save cycle time when event happened
    mov [DspEventBuffer+edi+4],edx  ;save register address and value
    add edi,byte 8              ;next event slot
    and edi,DspEventBufferSize-1  ;wrap around
    mov [DspBufferPointer],edi  ;save position in event buffer
    mov edx,0F3h                ;restore DSP data address
    ret

;====================
; (edx=dsp register to get value from, edi=buffer ptr, ebp=cycle time)
.RecordDspFrom:
    push ebx
    mov bh,[DspRam+edx]
    cmp [DspRamCopy+edx],bh
    je .RecordValueSame
    mov [DspRamCopy+edx],bh
    mov bl,dl                       ;copy register address
    mov [DspEventBuffer+edi],ebp    ;save cycle time when event happened
    mov [DspEventBuffer+edi+4],ebx  ;save both register address and value
    add edi,byte 8              ;next event slot
    and edi,DspEventBufferSize-1  ;wrap around
.RecordValueSame:
    pop ebx
    ret

;==============================
section .data

align 4
.ReadRegJumpTable:

.WriteRegJumpTable:
dd .wReg00,.wReg01,.wReg02,.wReg03,.wReg04,.wReg05,.wReg06,.wReg07,.wReg08,.wReg09,.wReg0A,.wReg0B,.wReg0C,.wReg0D,.wReg0E,.wReg0F
dd .wReg00,.wReg01,.wReg02,.wReg03,.wReg04,.wReg05,.wReg06,.wReg07,.wReg08,.wReg09,.wReg0A,.wReg0B,.wReg0C,.wReg0D,.wReg0E,.wReg0F
dd .wReg00,.wReg01,.wReg02,.wReg03,.wReg04,.wReg05,.wReg06,.wReg07,.wReg08,.wReg09,.wReg0A,.wReg0B,.wReg0C,.wReg0D,.wReg0E,.wReg0F
dd .wReg00,.wReg01,.wReg02,.wReg03,.wReg04,.wReg05,.wReg06,.wReg07,.wReg08,.wReg09,.wReg0A,.wReg0B,.wReg0C,.wReg0D,.wReg0E,.wReg0F
dd .wReg00,.wReg01,.wReg02,.wReg03,.wReg04,.wReg05,.wReg06,.wReg07,.wReg08,.wReg09,.wReg0A,.wReg0B,.wReg4C,.wReg0D,.wReg0E,.wReg0F
dd .wReg00,.wReg01,.wReg02,.wReg03,.wReg04,.wReg05,.wReg06,.wReg07,.wReg08,.wReg09,.wReg0A,.wReg0B,.wReg5C,.wReg0D,.wReg0E,.wReg0F
dd .wReg00,.wReg01,.wReg02,.wReg03,.wReg04,.wReg05,.wReg06,.wReg07,.wReg08,.wReg09,.wReg0A,.wReg0B,.wReg0C,.wReg0D,.wReg0E,.wReg0F
dd .wReg00,.wReg01,.wReg02,.wReg03,.wReg04,.wReg05,.wReg06,.wReg07,.wReg08,.wReg09,.wReg0A,.wReg0B,.wReg7C,.wReg0D,.wReg0E,.wReg0F
dd .wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX
dd .wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX
dd .wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX
dd .wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX
dd .wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX
dd .wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX
dd .wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX
dd .wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX,.wRegXX

SECTION .text

;=============================================================================
; Expands the rle compressed pitch table, from 256 words to 16k
; Each pitch is converted to the nearest half-semitone, which after adding
; to the current transposition and dividing by two will yield the nearest
; whole semitone. It then procedes to DspEmuReinit
DspEmuInit:
    cld
    mov esi,PitchTableRle
    xor ecx,ecx             ;zero extend counter
    mov edi,PitchTable
    xor eax,eax             ;zero current MIDI note
.NextNote:
    mov cx,[esi]            ;get RLE length
    rep stosb               ;fill table with current note
    add esi,byte 2          ;next strip
    inc al                  ;next MIDI note
    jnz .NextNote           ;stop on 256 wrap

; Makes a copy of DSP RAM, clears sample buffer
DspEmuReinit:
    ;Make copy of DSP RAM
    cld
    mov esi,DspRam
    mov edi,DspRamCopy
    mov ecx,128/4
    rep movsd

  %if (SamplesBuffered+SamplesBuffered_Size != SampleInfoTable)
    %error SamplesInfoTable must come immediately after SamplesBuffered
  %endif
    ;Mark all memory as being noncached (no BRR expansion yet)
    ;all samples as being unused
    ;all information to zero
    mov edi,SamplesBuffered
    mov ecx,(SamplesBuffered_Size + SampleInfoTable_Size)/4
    xor eax,eax
    rep stosd

;    xor ecx,ecx
;.NextIns:
;    mov [SampleInfoTable.Instrument+ecx],cl
;    inc cl
;    jnz .NextIns
    ret

;==============================
BufferAllSamples:
    mov esi,Text.BufferingSamples
    call AddMessage

    movzx ebx,byte [DspRam+DspReg.SourceDirectory]
    shl ebx,8                   ;*256
    add ebx,SpcRam
    push dword 0                ;set starting sample
.Next:
    push ebx                    ;save source directory ptr
    mov esi,[ebx]               ;get sample offset from table
    push esi                    ;save sample offsets
    and esi,65535               ;bottom word only for sample
    push esi
    call BrrSampleExpand
    mov [.Loop],al              ;save header byte
    pop esi
    push edx                    ;save sample length
    add esi,SpcRam              ;GetCheckSum requires an absolute address
    call GetCheckSum
    push dword 0                ;set loop length in case jump is taken
    test byte [.Loop],2         ;check if sample is looped
    jz .NoSampleLoop

    mov esi,[esp+8]             ;retrieve sample offset
    push eax                    ;save checksum
    shr esi,16                  ;top word only for loop portion
    push esi
    call BrrSampleExpand
    pop esi
    pop eax                     ;get previous checksum of sample
    mov [esp],edx               ;save sample length of loop portion
    add esi,SpcRam              ;GetCheckSum requires an absolute address
    call GetCheckSum.StartingWith
.NoSampleLoop:

    mov ecx,[esp+16]            ;get sample number
    mov [SampleInfoTable.Checksum+ecx*4],eax    ;save checksum of both sample and loop portion
    pop dword [SampleInfoTable.LoopLen+ecx*4]   ;set sample loop length in table
    pop dword [SampleInfoTable.Length+ecx*4]    ;set sample length in table
    pop dword [SampleInfoTable.Offset+ecx*4]    ;set offsets of sample and loop portion
    pop ebx
    add ebx,byte 4
    inc byte [esp]
    jnz .Next

    pop ecx
    ret

section .data
.Loop:  db 0
section .text

;==============================
; Expands a BRR sample into the buffer.
; Returns only the length of the sample given. For the length of the loop
; portion too, the loop portion offset in RAM must be passed to it separately
; then the length of the original sample and loop portion can be added
; together to get the total sample length. If sample starts in direct page
; RAM (highly unlikely), starts in the ROM (even more unlikely), or is not
; at least three blocks long, then the sample is disregarded and returns zero.
;
; (esi=offset in SPC RAM) (ecx=byte length, edx=sample length, eax=last header byte, esi=byte just after last sample)
BrrSampleExpand.Invalid:
    xor eax,eax                 ;set header byte to zero
    xor ecx,ecx                 ;set both byte and sample length to 0
    xor edx,edx
    ret

BrrSampleExpand:
    cmp esi,256                 ;sample should not start in direct page RAM
    jb .Invalid
    cmp esi,65463               ;sample should not start in ROM
    ja .Invalid
    mov al,[SpcRam+esi]         ;get first header byte
    mov ah,[SpcRam+esi+9]       ;get second header byte
    test eax,0101h              ;sample must be at least 3 blocks long (48 samples)
    jnz .Invalid                ;or else consider it all garbage
    cmp al,12<<4                ;range must be valid
    ja .Invalid                 ;do samples sometimes contain invalid ranges???
    cmp ah,12<<4                ;second range must also be valid
    ja .Invalid

; This can be called when the source is without doubt a valid sample to skip
; the unnecessary checks, or to skip them when an unlikely sample actually is
; valid.
.IgnoreChecks:
    ;calculate desination in buffer (sample_offset / 9 * 32)
    mov eax,esi
    mov ebx,9
    xor edx,edx
    div ebx                     ;/9
    push dword 0                ;set sample length counter to zero
    shl eax,5                   ;*32
    bts [SamplesBuffered],esi   ;mark sample as cached
    lea edi,[SamplesBuffer+eax]
.Next:
    mov al,[SpcRam+esi]         ;get header byte
    mov ch,8                    ;set counter for eight samples to follow
    inc esi
    mov cl,al                   ;copy header byte
    shr cl,4                    ;get range from top four bits
.NextByte:
    movsx edx,byte [SpcRam+esi] ;get sign extended high nybble
    mov ebx,edx
    shl edx,28                  ;move low nybble to top of dword
    shr ebx,4                   ;move high nybble down four bits
    sar edx,28                  ;sign extend low nybble
    shl ebx,cl                  ;apply range to high nybble
    shl edx,cl                  ;apply range to low nybble
    mov [edi],bx                ;high nybble is stored
    mov [edi+2],dx              ;before low nybble
    inc esi
    add edi,byte 4
    ;cmp edi,SamplesBuffer+SamplesBuffer_Size-4;!!!temp safety check
    ;jbe .DestOk
    ;int3
  ;.DestOk:
    dec ch
    jnz .NextByte
    add dword [esp],byte 16
    test eax,1
    jnz .BrrEnd
    cmp esi,65463
    jbe .Next
.BrrEnd:

    pop edx                     ;retrieve length in samples
    mov ecx,edx
    shr ecx,4                   ;     byte_length = sample_length / 16 * 9
    lea ecx,[ecx*8+ecx]         ;[OR] byte_length = starting_offset - ending_offset
    ret

;==============================
; (esi=source, ecx=byte length) (eax=addition checksum)
;
; This function's purpose is for uniquely identifying a given sample,
; although it could actually checksum anything. The checksum that it returns
; is the dword sum of the raw sample, not the expanded sample. The final
; dword (for any samples with lengths not an even multiple of four) will be
; masked to exclude the last few bytes which are not part of the sample, then
; added to the checksum. For a complete checksum of a sample, this function
; must be called twice, once for the sample beginning and again for the loop
; portion (if it has a loop portion).
GetCheckSum:
    xor eax,eax                 ;zero checksum
;(eax=existing checksum)
.StartingWith:
    mov edx,ecx                 ;copy length
    shr ecx,2                   ;/4 get dwords
    jz .LastBytes
.NextDword:
    add eax,[esi]               ;add to checksum
    add esi,byte 4
    loop .NextDword
.LastBytes:
    and edx,3                   ;%4 get leftover bytes
    shl edx,3                   ;*8
    bts ecx,edx                 ;set bit in ecx (which is zero from the loop above)
    dec ecx                     ;make mask, = 1 << ((sampleLength and 3) * 8)
    and ecx,[esi]               ;get final dword
    add eax,ecx                 ;add final few bytes to checksum
    ret

;==============================
OutputDspWrites:
    mov edi,[DspBufferPointer]
    mov esi,[DspBufferWritten]
    mov [DspBufferWritten],edi
    cmp edi,esi
    jz .End
    sub edi,esi
    and edi,DspEventBufferSize-1
    mov ebx,[DspFileHandle]
    mov ecx,6
.Next:
    mov ah,40h                  ;DOS: Write to file
    lea edx,[DspEventBuffer+esi]
    int 21h
    add esi,byte 8
    and esi,DspEventBufferSize-1
    sub edi,byte 8
    ja .Next
.End:
    ret

;==============================
NoteBufferSeekNext:
    mov esi,[NoteBufferPos]
    mov eax,[CycleCount]
    and esi,NoteBufferSize-1
    mov edx,[NoteBufferEnd]
    mov [NoteBufferPrevPos],esi
    jmp short .First
.Next:
    cmp [NoteBuffer+esi],eax    ;check cycle time when note was recorded with current time
    ja .End
    add esi,byte 8
    and esi,NoteBufferSize-1
.First:
    cmp esi,edx                 ;check current position with end
    jne .Next
.End:
    mov [NoteBufferPos],esi
    ret

;==============================
DumpSoundBuffer:
    cld
    mov edi,SamplesBuffer
    mov ecx,SamplesBuffer_Size/4
    xor eax,eax
    rep stosd
    call BufferAllSamples
    mov ah,3Ch                  ;DOS: Create file
    xor ecx,ecx                 ;no attributes
    mov edx,SoundDumpFileName
    int 21h
    jc .FileError
    mov ebx,eax
    mov edx,SoundDumpHeader
    mov ecx,SoundDumpHeader_Size
    mov ah,40h                  ;DOS: Write to file
    int 21h
    mov edx,SamplesBuffer
    mov ecx,SamplesBuffer_Size
    mov ah,40h                  ;DOS: Write to file
    int 21h
    mov ah,3Eh                  ;DOS: Close file
    int 21h
.FileError:
    ret

.Info:
    mov edx,Text.SampleInfoTitles
    call WriteString
    xor ecx,ecx
.NextSampleInfoLine:
    push ecx
    push dword [SampleInfoTable.Checksum+ecx*4]
    push dword [SampleInfoTable.LoopLen+ecx*4]
    push dword [SampleInfoTable.Length+ecx*4]
    push dword [SampleInfoTable.Offset+ecx*4]
    mov eax,ecx
    mov edi,Text.SampleInfo
    mov ecx,3
    call NumToString.OfLength
    movzx eax,word [esp]
    mov edi,Text.SampleInfo+8
    mov ecx,5
    call NumToString.OfLength
    pop eax
    shr eax,16
    mov edi,Text.SampleInfo+14
    mov ecx,5
    call NumToString.OfLength
    pop eax
    mov edi,Text.SampleInfo+21
    mov ecx,5
    call NumToString.OfLength
    pop eax
    mov edi,Text.SampleInfo+27
    mov ecx,5
    call NumToString.OfLength
    pop eax
    mov edi,Text.SampleInfo+34
    mov ecx,32                  ;bits
    call MakeHexNum
    mov edx,Text.SampleInfo
    call WriteString
    pop ecx
    inc ecx
    cmp ecx,255
    jbe near .NextSampleInfoLine
    ret

;==============================
GetSampleWaveAttributes:
; My dream function that awaits completion! :)
; Not sure exactly how I'll get it to work... :(
;
; Scans from the start point to the end of the sample (usually starting at
; the loop point or beginning of the sample if there is none) to return the
; dominant frequency and average volume. Also returns the nearest
; half-semitone offset from A440.
;
; The algorithm uses multiple sliding windows, one for each possible
; frequency. They all move in parallel from start to end, accumulating the
; differences between the averages of two areas, the left and right halves.
; The closer to zero that the accumulated total of a particular frequency is,
; the closer to cancellation. The higher the total, the more resonantly it
; matches. Then each result is correctly weighted to slightly favor higher
; frequencies (which the ear generally hears more predominantly than low
; tones) by multiplication. Finally the highest frequency, or highest between
; two frequencies, is selected and returned.
;
; To see the example below, it must be viewed it MS-DOS Edit or another ASCII
; text editor.
;
;  1                  Ŀ    Ŀ
;  0    
; -1                           
;    L R    LEFT - RIGHT = TOTAL
;    ٳ        4 - -2 =  6    2+6+8 =  16  close
;     ٳ        4 - -4 =  8
;      ٳ        2 - -4 =  6
;       ٳ        0 - -2 =  2
;        ٳ       -2 -  0 = -2
;                -4 -  2 = -6
;               5 - -5 = 10    2+6+10 = 18  perfect resonance
;               3 - -3 =  6
;               1 - -1 =  2
;              -1 -  1 = -2
;              -3 -  3 = -6
;              -5 -  5 = 10
;             4 - -2 =  6    2+6+8 =  16  close
;             2 -  0 =  2
;             0 - -2 = -2
;            -2 -  4 = -6
;            -4 -  4 = -8
;            -4 -  2 = -6
;     0 -  0 =  0    0+0+0 = 0    complete cancellation
    ;ret

;*****************************************************************************
%if 0
    jmp short .LoopStart
.LoopTop:
    add esi,byte 4
    inc edi
.LoopStart:
    mov eax,[esi]
    cmp eax,32767
    jbe .SampleOk       ;if sample is 0 to 32767, skip straight to writing it
    cmp eax,-32768
    jae .SampleOk       ;if sample is -32768 to -1, skip straight to writing it
    sar eax,23          ;extend sign bit down
    not eax
    mov [es:edi],ah
    dec ecx
    jnz .Loop
    jmp short .WasOverflow
.SampleOk:
    xor eax,8000h       ;invert sign  -128=0, 0=128, 1=129, 127=255
    mov [es:edi],ah
    dec ecx
    jnz .Loop

.LoopTop:
    add esi,byte 4
    inc edi
.LoopStart:
    mov eax,[esi]
    cmp eax,-32768
    ja .NoOverflow
    test eax,-1^255
    jz .NoOverflow
    seto ah
    dec ah
    mov [es:edi],ah
    dec ecx
    jnz .LoopTop
    jmp short .WasOverflow
.NoOverflow:
    xor ah,80h
    mov [es:edi],ah
    dec ecx
    jnz .LoopTop
.WasOverflow:

%endif

%if 0
on/off/slide    action to do to note, press, release, or slide pitch
source          sample number
note number     half-semitone (after pitch mapping, before transposition)
volume          before volume multiplication

.KeyOn:
    mov [NoteBuffer+edi],4Ch                        ;set action
    movzx ebx,byte [DspRam+esi+DspReg.SourceNumber] ;get sample number
    mov [NoteBuffer+edi+1],bl                       ;set sample
    movzx eax,word [DspRam+esi+DspReg.PitchLow]     ;get pitch
    cmp ax,[HighestPitches+ebx*2]
    jbe .NotHighestPitch
    mov [HighestPitch+ebx*2],ax
.NotHighestPitch:
    imul dword [SamplePitchRatios+ebx*4]            ;multiply by ratio
    shrd eax,edx,16             ;divide by 65536
    cmp eax,16384               ;pitch must 0-16383 (14bit)
    jb .PitchOk
    mov eax,16383
  .PitchOk:
    mov dl,[PitchTable+eax]     ;get MIDI note number
    mov [NoteBuffer+edi+2],dl   ;set note number

    mov al,[DspRam+DspReg.VolLeft]
    mov [NoteBuffer+edi+3],al   ;set volume

;    add al,ah                   ;combine channnels (!!! 128+128 > 255)
;    test al,al                  ;check for sign
;    jns .VolumeNotSigned
;    neg al                      ;make positive
;  .VolumeNotSigned:
;    test ah,ah                  ;check for sign
;    jns .RightVolumeNotSigned
;    neg ah                      ;make positive
;  .RightVolumeNotSigned:
;    add al,ah                   ;they should both be positive now
;    sbb al,0                    ;but just in case of overflow
;    shr al,1                    ;average the two channels
;    rcr al,1    127
%endif
